# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for forward dynamics functions."""

import mujoco
import numpy as np
import warp as wp
from absl.testing import absltest
from absl.testing import parameterized

import mujoco_warp as mjw
from mujoco_warp import BiasType
from mujoco_warp import DisableBit
from mujoco_warp import EnableBit
from mujoco_warp import GainType
from mujoco_warp import IntegratorType
from mujoco_warp import test_data

# tolerance for difference between MuJoCo and mjwarp smooth calculations - mostly
# due to float precision
_TOLERANCE = 5e-5


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 10  # avoid test noise
  err_msg = f"mismatch: {name}"
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


class ForwardTest(parameterized.TestCase):
  # TODO(team): test sparse when actuator_moment and/or ten_J have sparse representation
  @parameterized.product(xml=["humanoid/humanoid.xml", "pendula.xml"])
  def test_fwd_velocity(self, xml):
    _, mjd, m, d = test_data.fixture(xml, qvel_noise=0.01, ctrl_noise=0.1)

    for arr in (d.actuator_velocity, d.qfrc_bias):
      arr.zero_()

    mjw.fwd_velocity(m, d)

    _assert_eq(d.actuator_velocity.numpy()[0], mjd.actuator_velocity, "actuator_velocity")
    _assert_eq(d.qfrc_bias.numpy()[0], mjd.qfrc_bias, "qfrc_bias")

  def test_fwd_velocity_tendon(self):
    _, mjd, m, d = test_data.fixture("tendon/fixed.xml")

    d.ten_velocity.zero_()
    mjw.fwd_velocity(m, d)

    _assert_eq(d.ten_velocity.numpy()[0], mjd.ten_velocity, "ten_velocity")

  @parameterized.product(
    xml=("actuation/actuation.xml", "actuation/actuators.xml", "actuation/muscle.xml"),
    disableflags=(0, DisableBit.ACTUATION),
  )
  def test_actuation(self, xml, disableflags):
    mjm, mjd, m, d = test_data.fixture(xml, keyframe=0, overrides={"opt.disableflags": disableflags})

    for arr in (d.qfrc_actuator, d.actuator_force, d.act_dot):
      arr.zero_()

    mjw.fwd_actuation(m, d)

    _assert_eq(d.qfrc_actuator.numpy()[0], mjd.qfrc_actuator, "qfrc_actuator")
    _assert_eq(d.actuator_force.numpy()[0], mjd.actuator_force, "actuator_force")

    if mjm.na:
      _assert_eq(d.act_dot.numpy()[0], mjd.act_dot, "act_dot")

      # next activations
      mujoco.mj_step(mjm, mjd)
      mjw.step(m, d)

      _assert_eq(d.act.numpy()[0], mjd.act, "act")

    # TODO(team): test actearly

  @parameterized.parameters(0, DisableBit.CLAMPCTRL)
  def test_clampctrl(self, disableflags):
    _, mjd, _, d = test_data.fixture(
      xml="""
    <mujoco>
      <worldbody>
        <body>
          <joint name="joint" type="slide"/>
          <geom type="sphere" size=".1"/>
        </body>
      </worldbody>
      <actuator>
        <motor joint="joint" ctrlrange="-1 1"/>
      </actuator>
      <keyframe>
        <key ctrl="2"/>
      </keyframe>
    </mujoco>
    """,
      keyframe=0,
      overrides={"opt.disableflags": disableflags},
    )

    _assert_eq(d.ctrl.numpy()[0], mjd.ctrl, "ctrl")

  def test_fwd_acceleration(self):
    _, mjd, m, d = test_data.fixture("humanoid/humanoid.xml", qvel_noise=0.01, ctrl_noise=0.1)

    for arr in (d.qfrc_smooth, d.qacc_smooth):
      arr.zero_()

    mjw.fwd_acceleration(m, d)

    _assert_eq(d.qfrc_smooth.numpy()[0], mjd.qfrc_smooth, "qfrc_smooth")
    _assert_eq(d.qacc_smooth.numpy()[0], mjd.qacc_smooth, "qacc_smooth")

  @parameterized.product(
    jacobian=(mujoco.mjtJacobian.mjJAC_AUTO, mujoco.mjtJacobian.mjJAC_DENSE), disableflags=(0, DisableBit.EULERDAMP)
  )
  def test_euler(self, jacobian, disableflags):
    mjm, mjd, _, _ = test_data.fixture(
      "pendula.xml",
      qvel_noise=0.01,
      ctrl_noise=0.1,
      overrides={"opt.jacobian": jacobian, "opt.disableflags": DisableBit.CONTACT | disableflags},
    )
    self.assertTrue((mjm.dof_damping > 0).any())

    m = mjw.put_model(mjm)
    d = mjw.put_data(mjm, mjd)
    mujoco.mj_forward(mjm, mjd)
    mujoco.mj_Euler(mjm, mjd)
    mjw.solve(m, d)  # compute efc.Ma
    mjw.euler(m, d)

    _assert_eq(d.qpos.numpy()[0], mjd.qpos, "qpos")
    _assert_eq(d.qvel.numpy()[0], mjd.qvel, "qvel")
    _assert_eq(d.act.numpy()[0], mjd.act, "act")

  def test_rungekutta4(self):
    mjm, mjd, m, d = test_data.fixture(
      xml="""
        <mujoco>
          <option integrator="RK4" iterations="1" ls_iterations="1">
            <flag constraint="disable"/>
          </option>
          <worldbody>
            <body>
              <joint type="hinge"/>
              <geom type="sphere" size=".1"/>
              <body pos="0.1 0 0">
                <joint type="hinge"/>
                <geom type="sphere" size=".1"/>
              </body>
            </body>
          </worldbody>
          <keyframe>
            <key qpos=".1 .2" qvel=".025 .05"/>
          </keyframe>
        </mujoco>
        """,
      keyframe=0,
    )

    mjw.rungekutta4(m, d)
    mujoco.mj_RungeKutta(mjm, mjd, 4)

    _assert_eq(d.qpos.numpy()[0], mjd.qpos, "qpos")
    _assert_eq(d.qvel.numpy()[0], mjd.qvel, "qvel")
    _assert_eq(d.time.numpy()[0], mjd.time, "time")
    _assert_eq(d.xpos.numpy()[0], mjd.xpos, "xpos")

    # test rungekutta determinism
    def rk_step() -> wp.array(dtype=wp.float32, ndim=2):
      d.qpos = wp.ones_like(d.qpos)
      d.qvel = wp.ones_like(d.qvel)
      d.act = wp.ones_like(d.act)
      mjw.rungekutta4(m, d)
      return d.qpos

    _assert_eq(rk_step().numpy()[0], rk_step().numpy()[0], "qpos")

  @parameterized.product(
    jacobian=(mujoco.mjtJacobian.mjJAC_SPARSE, mujoco.mjtJacobian.mjJAC_DENSE),
    actuation=(0, DisableBit.ACTUATION),
    spring=(0, DisableBit.SPRING),
    damper=(0, DisableBit.DAMPER),
  )
  def test_implicit(self, jacobian, actuation, spring, damper):
    mjm, mjd, _, _ = test_data.fixture(
      "pendula.xml",
      overrides={
        "opt.jacobian": jacobian,
        "opt.disableflags": DisableBit.CONTACT | actuation | spring | damper,
        "opt.integrator": IntegratorType.IMPLICITFAST,
      },
    )

    mjm.actuator_gainprm[:, 2] = np.random.uniform(low=0.01, high=10.0, size=mjm.actuator_gainprm[:, 2].shape)

    # change actuators to velocity/damper to cover all codepaths
    mjm.actuator_gaintype[3] = GainType.AFFINE
    mjm.actuator_gaintype[6] = GainType.AFFINE
    mjm.actuator_biastype[0:3] = BiasType.AFFINE
    mjm.actuator_biastype[4:6] = BiasType.AFFINE
    mjm.actuator_biasprm[0:3, 2] = -1.0
    mjm.actuator_biasprm[4:6, 2] = -1.0
    mjm.actuator_ctrlrange[3:7] = 10.0
    mjm.actuator_gear[:] = 1.0

    mjd.qvel = np.random.uniform(low=-0.01, high=0.01, size=mjd.qvel.shape)
    mjd.ctrl = np.random.uniform(low=-0.1, high=0.1, size=mjd.ctrl.shape)
    mjd.act = np.random.uniform(low=-0.1, high=0.1, size=mjd.act.shape)
    mujoco.mj_forward(mjm, mjd)

    m = mjw.put_model(mjm)
    d = mjw.put_data(mjm, mjd)
    # compute efc.Ma - used by mjw.implicit
    d.efc.Ma = wp.array(mjd.qfrc_constraint + mjd.qfrc_smooth, dtype=wp.float32, shape=(1, -1))

    mujoco.mj_implicit(mjm, mjd)
    mjw.implicit(m, d)

    _assert_eq(d.qpos.numpy()[0], mjd.qpos, "qpos")
    _assert_eq(d.act.numpy()[0], mjd.act, "act")

  def test_implicit_position(self):
    mjm, mjd, m, d = test_data.fixture(
      "actuation/position.xml",
      keyframe=0,
      qvel_noise=0.01,
      ctrl_noise=0.1,
      overrides={"opt.integrator": IntegratorType.IMPLICITFAST},
    )

    mujoco.mj_implicit(mjm, mjd)

    mjw.solve(m, d)  # compute efc.Ma
    mjw.implicit(m, d)

    _assert_eq(d.qpos.numpy()[0], mjd.qpos, "qpos")
    _assert_eq(d.qvel.numpy()[0], mjd.qvel, "qvel")

  def test_implicit_tendon_damping(self):
    mjm, mjd, m, d = test_data.fixture(
      "tendon/damping.xml",
      keyframe=0,
      qvel_noise=0.01,
      ctrl_noise=0.1,
      overrides={"opt.integrator": IntegratorType.IMPLICITFAST},
    )

    mujoco.mj_implicit(mjm, mjd)

    mjw.solve(m, d)  # compute efc.Ma
    mjw.implicit(m, d)

    _assert_eq(d.qpos.numpy()[0], mjd.qpos, "qpos")
    _assert_eq(d.qvel.numpy()[0], mjd.qvel, "qvel")

  @absltest.skipIf(not wp.get_device().is_cuda, "Skipping test that requires GPU.")
  @parameterized.product(
    xml=("humanoid/humanoid.xml", "pendula.xml", "constraints.xml", "collision.xml"), graph_conditional=(True, False)
  )
  def test_graph_capture(self, xml, graph_conditional):
    # TODO(team): test more environments
    _, _, m, d = test_data.fixture(xml, overrides={"opt.graph_conditional": graph_conditional})

    with wp.ScopedCapture() as capture:
      mjw.step(m, d)

    # step a few times to ensure no errors at the step boundary
    wp.capture_launch(capture.graph)
    wp.capture_launch(capture.graph)
    wp.capture_launch(capture.graph)

    self.assertTrue(d.time.numpy()[0] > 0.0)

  def test_forward_energy(self):
    _, mjd, _, d = test_data.fixture(
      "humanoid/humanoid.xml", qvel_noise=0.01, ctrl_noise=0.1, overrides={"opt.enableflags": EnableBit.ENERGY}
    )

    _assert_eq(d.energy.numpy()[0][0], mjd.energy[0], "potential energy")
    _assert_eq(d.energy.numpy()[0][1], mjd.energy[1], "kinetic energy")

  def test_tendon_actuator_force_limits(self):
    for keyframe in range(7):
      _, mjd, m, d = test_data.fixture("actuation/tendon_force_limit.xml", keyframe=keyframe)

      d.actuator_force.zero_()

      mjw.forward(m, d)

      _assert_eq(d.actuator_force.numpy()[0], mjd.actuator_force, "actuator_force")

  @parameterized.product(xml=("humanoid/humanoid.xml",), energy=(0, EnableBit.ENERGY))
  def test_step1(self, xml, energy):
    # TODO(team): test more mjcfs
    mjm, mjd, m, d = test_data.fixture(
      xml, qpos_noise=0.1, qvel_noise=0.01, ctrl_noise=0.1, overrides={"opt.enableflags": energy}
    )

    # some of the fields updated by step1
    step1_field = [
      "xpos",
      "xquat",
      "xmat",
      "xipos",
      "ximat",
      "xanchor",
      "xaxis",
      "geom_xpos",
      "geom_xmat",
      "site_xmat",
      "subtree_com",
      "cinert",
      "cdof",
      "cam_xpos",
      "cam_xmat",
      "light_xpos",
      "light_xdir",
      "ten_length",
      "ten_J",
      "ten_wrapadr",
      "ten_wrapnum",
      "wrap_obj",
      "wrap_xpos",
      "qM",
      "qLD",
      "nefc",
      "efc_type",
      "efc_id",
      "efc_J",
      "efc_pos",
      "efc_margin",
      "efc_D",
      "efc_vel",
      "efc_aref",
      "efc_frictionloss",
      "actuator_length",
      "actuator_moment",
      "actuator_velocity",
      "ten_velocity",
      "cvel",
      "cdof_dot",
      "qfrc_spring",
      "qfrc_damper",
      "qfrc_gravcomp",
      "qfrc_fluid",
      "qfrc_passive",
      "qfrc_bias",
      "energy",
    ]
    if m.nflexvert:
      step1_field += ["flexvert_xpos"]
    if m.nflexedge:
      step1_field += ["flexedge_length", "flexedge_velocity"]

    def _getattr(arr):
      if (len(arr) >= 4) & (arr[:4] == "efc_"):
        return getattr(d.efc, arr[4:]), True
      return getattr(d, arr), False

    for arr in step1_field:
      if arr in ("geom_xpos", "geom_xmat"):
        # leave geom_xpos and geom_xmat untouched because they have static data
        continue
      attr, _ = _getattr(arr)
      if arr in ("xquat", "xmat", "ximat"):
        # xquat, xmat, ximat need to retain identity for world body
        attr = attr[:, 1:]
      if attr.dtype == float:
        attr.fill_(wp.nan)
      elif attr.dtype == int:
        attr.fill_(-1)
      else:
        attr.zero_()

    mujoco.mj_step1(mjm, mjd)
    mjw.step1(m, d)

    # Precompute sorting for efc fields to avoid non determinism
    nefc = d.nefc.numpy()[0]
    if nefc > 0:
      nv = m.nv
      d_efc_J = d.efc.J.numpy()[0, :nefc, :nv]
      if mjd.efc_J.shape[0] != mjd.nefc * mjm.nv:
        mjd_efc_J = np.zeros((mjd.nefc, mjm.nv))
        mujoco.mju_sparse2dense(mjd_efc_J, mjd.efc_J, mjd.efc_J_rownnz, mjd.efc_J_rowadr, mjd.efc_J_colind)
      else:
        mjd_efc_J = mjd.efc_J.reshape((mjd.nefc, mjm.nv))

      # Sort by efc_type, then efc_J columns (tiebreaker)
      d_sort = np.lexsort((*d_efc_J.T, d.efc.type.numpy()[0, :nefc]))
      mjd_sort = np.lexsort((*mjd_efc_J.T, mjd.efc_type[:nefc]))
      _assert_eq(d_efc_J[d_sort].reshape(-1), mjd_efc_J[mjd_sort].reshape(-1), "efc_J")

      # Check efc_id here as a contact may have a different id
      _assert_eq(sorted(d.efc.id.numpy()[0, :nefc]), sorted(mjd.efc_id[:nefc]), "efc_id")

    for arr in step1_field:
      d_arr, is_efc = _getattr(arr)

      d_arr = d_arr.numpy()[0]
      mjd_arr = getattr(mjd, arr)
      if arr in ["xmat", "ximat", "geom_xmat", "site_xmat", "cam_xmat"]:
        mjd_arr = mjd_arr.reshape(-1)
        d_arr = d_arr.reshape(-1)
      elif arr == "qM":
        qM = np.zeros((mjm.nv, mjm.nv))
        mujoco.mj_fullM(mjm, qM, mjd.qM)
        mjd_arr = qM
        d_arr = d_arr[: mjm.nv, : mjm.nv]
      elif arr == "actuator_moment":
        actuator_moment = np.zeros((mjm.nu, mjm.nv))
        mujoco.mju_sparse2dense(actuator_moment, mjd.actuator_moment, mjd.moment_rownnz, mjd.moment_rowadr, mjd.moment_colind)
        mjd_arr = actuator_moment
      elif arr == "ten_J" and mjm.ntendon:
        ten_J = np.zeros((mjm.ntendon, mjm.nv))
        mujoco.mju_sparse2dense(ten_J, mjd.ten_J, mjd.ten_J_rownnz, mjd.ten_J_rowadr, mjd.ten_J_colind)
        mjd_arr = ten_J
      elif arr == "efc_J" or arr == "efc_id":
        # Already checked earlier
        continue
      elif arr == "qLD":
        vec = np.ones((1, mjm.nv))
        res = np.zeros((1, mjm.nv))
        mujoco.mj_solveM(mjm, mjd, res, vec)

        vec_wp = wp.array(vec, dtype=float)
        res_wp = wp.zeros((1, mjm.nv), dtype=float)
        mjw.solve_m(m, d, res_wp, vec_wp)

        d_arr = res_wp.numpy()[0]
        mjd_arr = res[0]

      if is_efc:
        nefc = d.nefc.numpy()[0]
        nv = m.nv
        d_arr = d_arr[:nefc]
        d_arr = d_arr[d_sort].reshape(-1)
        mjd_arr = mjd_arr[mjd_sort].reshape(-1)

      _assert_eq(d_arr, mjd_arr, arr)

    # TODO(team): sensor_pos
    # TODO(team): sensor_vel

  @parameterized.product(
    xml=("humanoid/humanoid.xml",),
    integrator=(IntegratorType.EULER, IntegratorType.IMPLICITFAST, IntegratorType.RK4),
  )
  def test_step2(self, xml, integrator):
    mjm, mjd, m, _ = test_data.fixture(xml, qvel_noise=0.01, ctrl_noise=0.1, overrides={"opt.integrator": integrator})

    # some of the fields updated by step2
    step2_field = [
      "act_dot",
      "actuator_force",
      "qfrc_actuator",
      "qfrc_smooth",
      "qacc",
      "qacc_warmstart",
      "qvel",
      "qpos",
      "efc_force",
      "qfrc_constraint",
    ]

    def _getattr(arr):
      if (len(arr) >= 4) & (arr[:4] == "efc_"):
        return getattr(d.efc, arr[4:]), True
      return getattr(d, arr), False

    mujoco.mj_step1(mjm, mjd)

    # input
    ctrl = 0.1 * np.random.rand(mjm.nu)
    qfrc_applied = 0.1 * np.random.rand(mjm.nv)
    xfrc_applied = 0.1 * np.random.rand(mjm.nbody, 6)

    mjd.ctrl = ctrl
    mjd.qfrc_applied = qfrc_applied
    mjd.xfrc_applied = xfrc_applied

    d = mjw.put_data(mjm, mjd)

    for arr in step2_field:
      if arr in ["qpos", "qvel"]:
        continue
      attr, _ = _getattr(arr)
      if attr.dtype == float:
        attr.fill_(wp.nan)
      elif attr.dtype == int:
        attr.fill_(-1)
      else:
        attr.zero_()

    mujoco.mj_step2(mjm, mjd)
    mjw.step2(m, d)

    for arr in step2_field:
      d_arr, is_efc = _getattr(arr)
      d_arr = d_arr.numpy()[0]
      if is_efc:
        d_arr = d_arr[: d.nefc.numpy()[0]]
      _assert_eq(d_arr, getattr(mjd, arr), arr)

  def test_step_no_dofs(self):
    """Tests step with no degrees of freedom."""
    _, _, m, d = test_data.fixture(
      xml="""
    <mujoco>
      <body>
        <geom type="sphere" size="1"/>
      </body>
    </mujoco>
    """
    )
    mjw.step(m, d)
    self.assertGreater(d.time.numpy()[0], 0.0)


if __name__ == "__main__":
  wp.init()
  absltest.main()



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Any

import warp as wp

from mujoco_warp._src.collision_convex import convex_narrowphase
from mujoco_warp._src.collision_primitive import primitive_narrowphase
from mujoco_warp._src.collision_sdf import sdf_narrowphase
from mujoco_warp._src.io import BLEEDING_EDGE_MUJOCO
from mujoco_warp._src.math import upper_tri_index
from mujoco_warp._src.types import MJ_MAXVAL
from mujoco_warp._src.types import BroadphaseFilter
from mujoco_warp._src.types import BroadphaseType
from mujoco_warp._src.types import CollisionContext
from mujoco_warp._src.types import CollisionType
from mujoco_warp._src.types import Data
from mujoco_warp._src.types import DisableBit
from mujoco_warp._src.types import GeomType
from mujoco_warp._src.types import Model
from mujoco_warp._src.types import mat23
from mujoco_warp._src.types import mat63
from mujoco_warp._src.warp_util import cache_kernel
from mujoco_warp._src.warp_util import event_scope

wp.set_module_options({"enable_backward": False})

# Corresponding table to MuJoCo's mjCOLLISIONFUNC table in engine_collision_driver.c
MJ_COLLISION_TABLE = {
  (GeomType.PLANE, GeomType.SPHERE): CollisionType.PRIMITIVE,
  (GeomType.PLANE, GeomType.CAPSULE): CollisionType.PRIMITIVE,
  (GeomType.PLANE, GeomType.ELLIPSOID): CollisionType.PRIMITIVE,
  (GeomType.PLANE, GeomType.CYLINDER): CollisionType.PRIMITIVE,
  (GeomType.PLANE, GeomType.BOX): CollisionType.PRIMITIVE,
  (GeomType.PLANE, GeomType.MESH): CollisionType.PRIMITIVE,
  (GeomType.HFIELD, GeomType.SPHERE): CollisionType.CONVEX,
  (GeomType.HFIELD, GeomType.CAPSULE): CollisionType.CONVEX,
  (GeomType.HFIELD, GeomType.ELLIPSOID): CollisionType.CONVEX,
  (GeomType.HFIELD, GeomType.CYLINDER): CollisionType.CONVEX,
  (GeomType.HFIELD, GeomType.BOX): CollisionType.CONVEX,
  (GeomType.HFIELD, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.SPHERE, GeomType.SPHERE): CollisionType.PRIMITIVE,
  (GeomType.SPHERE, GeomType.CAPSULE): CollisionType.PRIMITIVE,
  (GeomType.SPHERE, GeomType.ELLIPSOID): CollisionType.CONVEX,
  (GeomType.SPHERE, GeomType.CYLINDER): CollisionType.PRIMITIVE,
  (GeomType.SPHERE, GeomType.BOX): CollisionType.PRIMITIVE,
  (GeomType.SPHERE, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.CAPSULE, GeomType.CAPSULE): CollisionType.PRIMITIVE,
  (GeomType.CAPSULE, GeomType.ELLIPSOID): CollisionType.CONVEX,
  (GeomType.CAPSULE, GeomType.CYLINDER): CollisionType.CONVEX,
  (GeomType.CAPSULE, GeomType.BOX): CollisionType.PRIMITIVE,
  (GeomType.CAPSULE, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.ELLIPSOID, GeomType.ELLIPSOID): CollisionType.CONVEX,
  (GeomType.ELLIPSOID, GeomType.CYLINDER): CollisionType.CONVEX,
  (GeomType.ELLIPSOID, GeomType.BOX): CollisionType.CONVEX,
  (GeomType.ELLIPSOID, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.CYLINDER, GeomType.CYLINDER): CollisionType.CONVEX,
  (GeomType.CYLINDER, GeomType.BOX): CollisionType.CONVEX,
  (GeomType.CYLINDER, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.BOX, GeomType.BOX): CollisionType.CONVEX,  # overwritten by NATIVECCD disable flag
  (GeomType.BOX, GeomType.MESH): CollisionType.CONVEX,
  (GeomType.MESH, GeomType.MESH): CollisionType.CONVEX,
}


def create_collision_context(naconmax: int) -> CollisionContext:
  """Create a CollisionContext with allocated arrays."""
  return CollisionContext(
    collision_pair=wp.empty(naconmax, dtype=wp.vec2i),
    collision_pairid=wp.empty(naconmax, dtype=wp.vec2i),
    collision_worldid=wp.empty(naconmax, dtype=int),
  )


@wp.kernel
def _zero_nacon_ncollision(
  # Data out:
  nacon_out: wp.array(dtype=int),
  ncollision_out: wp.array(dtype=int),
):
  ncollision_out[0] = 0
  nacon_out[0] = 0


@wp.func
def _plane_filter(
  size1: float, size2: float, margin1: float, margin2: float, xpos1: wp.vec3, xpos2: wp.vec3, xmat1: wp.mat33, xmat2: wp.mat33
) -> bool:
  if size1 == 0.0:
    # geom1 is a plane
    dist = wp.dot(xpos2 - xpos1, wp.vec3(xmat1[0, 2], xmat1[1, 2], xmat1[2, 2]))
    if BLEEDING_EDGE_MUJOCO:
      return dist <= size2 + margin1 + margin2
    else:
      return dist <= size2 + wp.max(margin1, margin2)
  elif size2 == 0.0:
    # geom2 is a plane
    dist = wp.dot(xpos1 - xpos2, wp.vec3(xmat2[0, 2], xmat2[1, 2], xmat2[2, 2]))
    if BLEEDING_EDGE_MUJOCO:
      return dist <= size1 + margin1 + margin2
    else:
      return dist <= size1 + wp.max(margin1, margin2)

  return True


@wp.func
def _sphere_filter(size1: float, size2: float, margin1: float, margin2: float, xpos1: wp.vec3, xpos2: wp.vec3) -> bool:
  if BLEEDING_EDGE_MUJOCO:
    bound = size1 + size2 + margin1 + margin2
  else:
    bound = size1 + size2 + wp.max(margin1, margin2)
  dif = xpos2 - xpos1
  dist_sq = wp.dot(dif, dif)
  return dist_sq <= bound * bound


# TODO(team): improve performance by precomputing bounding box
@wp.func
def _aabb_filter(
  # In:
  center1: wp.vec3,
  center2: wp.vec3,
  size1: wp.vec3,
  size2: wp.vec3,
  margin1: float,
  margin2: float,
  xpos1: wp.vec3,
  xpos2: wp.vec3,
  xmat1: wp.mat33,
  xmat2: wp.mat33,
) -> bool:
  """Axis aligned boxes collision.

  references: see Ericson, Real-time Collision Detection section 4.2.
              filterBox: filter contact based on global AABBs.
  """
  center1 = xmat1 @ center1 + xpos1
  center2 = xmat2 @ center2 + xpos2

  if BLEEDING_EDGE_MUJOCO:
    margin = margin1 + margin2
  else:
    margin = wp.max(margin1, margin2)

  max_x1 = -MJ_MAXVAL
  max_y1 = -MJ_MAXVAL
  max_z1 = -MJ_MAXVAL
  min_x1 = MJ_MAXVAL
  min_y1 = MJ_MAXVAL
  min_z1 = MJ_MAXVAL

  max_x2 = -MJ_MAXVAL
  max_y2 = -MJ_MAXVAL
  max_z2 = -MJ_MAXVAL
  min_x2 = MJ_MAXVAL
  min_y2 = MJ_MAXVAL
  min_z2 = MJ_MAXVAL

  sign = wp.vec2(-1.0, 1.0)

  for i in range(2):
    for j in range(2):
      for k in range(2):
        corner1 = wp.vec3(sign[i] * size1[0], sign[j] * size1[1], sign[k] * size1[2])
        pos1 = xmat1 @ corner1

        corner2 = wp.vec3(sign[i] * size2[0], sign[j] * size2[1], sign[k] * size2[2])
        pos2 = xmat2 @ corner2

        if pos1[0] > max_x1:
          max_x1 = pos1[0]

        if pos1[1] > max_y1:
          max_y1 = pos1[1]

        if pos1[2] > max_z1:
          max_z1 = pos1[2]

        if pos1[0] < min_x1:
          min_x1 = pos1[0]

        if pos1[1] < min_y1:
          min_y1 = pos1[1]

        if pos1[2] < min_z1:
          min_z1 = pos1[2]

        if pos2[0] > max_x2:
          max_x2 = pos2[0]

        if pos2[1] > max_y2:
          max_y2 = pos2[1]

        if pos2[2] > max_z2:
          max_z2 = pos2[2]

        if pos2[0] < min_x2:
          min_x2 = pos2[0]

        if pos2[1] < min_y2:
          min_y2 = pos2[1]

        if pos2[2] < min_z2:
          min_z2 = pos2[2]

  if center1[0] + max_x1 + margin < center2[0] + min_x2:
    return False
  if center1[1] + max_y1 + margin < center2[1] + min_y2:
    return False
  if center1[2] + max_z1 + margin < center2[2] + min_z2:
    return False
  if center2[0] + max_x2 + margin < center1[0] + min_x1:
    return False
  if center2[1] + max_y2 + margin < center1[1] + min_y1:
    return False
  if center2[2] + max_z2 + margin < center1[2] + min_z1:
    return False

  return True


# TODO(team): improve performance by precomputing bounding box
@wp.func
def _obb_filter(
  # In:
  center1: wp.vec3,
  center2: wp.vec3,
  size1: wp.vec3,
  size2: wp.vec3,
  margin1: float,
  margin2: float,
  xpos1: wp.vec3,
  xpos2: wp.vec3,
  xmat1: wp.mat33,
  xmat2: wp.mat33,
) -> bool:
  """Oriented bounding boxes collision (see Gottschalk et al.), see mj_collideOBB."""
  if BLEEDING_EDGE_MUJOCO:
    margin = margin1 + margin2
  else:
    margin = wp.max(margin1, margin2)

  xcenter = mat23()
  normal = mat63()
  proj = wp.vec2()
  radius = wp.vec2()

  # compute centers in local coordinates
  xcenter[0] = xmat1 @ center1 + xpos1
  xcenter[1] = xmat2 @ center2 + xpos2

  # compute normals in global coordinates
  normal[0] = wp.vec3(xmat1[0, 0], xmat1[1, 0], xmat1[2, 0])
  normal[1] = wp.vec3(xmat1[0, 1], xmat1[1, 1], xmat1[2, 1])
  normal[2] = wp.vec3(xmat1[0, 2], xmat1[1, 2], xmat1[2, 2])
  normal[3] = wp.vec3(xmat2[0, 0], xmat2[1, 0], xmat2[2, 0])
  normal[4] = wp.vec3(xmat2[0, 1], xmat2[1, 1], xmat2[2, 1])
  normal[5] = wp.vec3(xmat2[0, 2], xmat2[1, 2], xmat2[2, 2])

  # check intersections
  for j in range(2):
    for k in range(3):
      for i in range(2):
        proj[i] = wp.dot(xcenter[i], normal[3 * j + k])
        if i == 0:
          size = size1
        else:
          size = size2

        # fmt: off
        radius[i] = (
            wp.abs(size[0] * wp.dot(normal[3 * i + 0], normal[3 * j + k]))
          + wp.abs(size[1] * wp.dot(normal[3 * i + 1], normal[3 * j + k]))
          + wp.abs(size[2] * wp.dot(normal[3 * i + 2], normal[3 * j + k]))
        )
        # fmt: on
      if radius[0] + radius[1] + margin < wp.abs(proj[1] - proj[0]):
        return False

  return True


def _broadphase_filter(opt_broadphase_filter: int, ngeom_aabb: int, ngeom_rbound: int, ngeom_margin: int):
  @wp.func
  def func(
    # Model:
    geom_aabb: wp.array3d(dtype=wp.vec3),
    geom_rbound: wp.array2d(dtype=float),
    geom_margin: wp.array2d(dtype=float),
    # Data in:
    geom_xpos_in: wp.array2d(dtype=wp.vec3),
    geom_xmat_in: wp.array2d(dtype=wp.mat33),
    # In:
    geom1: int,
    geom2: int,
    worldid: int,
  ) -> bool:
    # 1: plane
    # 2: sphere
    # 4: aabb
    # 8: obb

    aabb_id = worldid % ngeom_aabb if wp.static(ngeom_aabb > 1) else 0
    center1, center2 = geom_aabb[aabb_id, geom1, 0], geom_aabb[aabb_id, geom2, 0]
    size1, size2 = geom_aabb[aabb_id, geom1, 1], geom_aabb[aabb_id, geom2, 1]

    rbound_id = worldid % ngeom_rbound if wp.static(ngeom_rbound > 1) else 0
    rbound1, rbound2 = geom_rbound[rbound_id, geom1], geom_rbound[rbound_id, geom2]
    margin_id = worldid % ngeom_margin if wp.static(ngeom_margin > 1) else 0
    margin1, margin2 = geom_margin[margin_id, geom1], geom_margin[margin_id, geom2]
    xpos1, xpos2 = geom_xpos_in[worldid, geom1], geom_xpos_in[worldid, geom2]
    xmat1, xmat2 = geom_xmat_in[worldid, geom1], geom_xmat_in[worldid, geom2]

    if rbound1 == 0.0 or rbound2 == 0.0:
      if wp.static(opt_broadphase_filter & BroadphaseFilter.PLANE):
        return _plane_filter(rbound1, rbound2, margin1, margin2, xpos1, xpos2, xmat1, xmat2)
    else:
      if wp.static(opt_broadphase_filter & BroadphaseFilter.SPHERE):
        if not _sphere_filter(rbound1, rbound2, margin1, margin2, xpos1, xpos2):
          return False
      if wp.static(opt_broadphase_filter & BroadphaseFilter.AABB):
        if not _aabb_filter(center1, center2, size1, size2, margin1, margin2, xpos1, xpos2, xmat1, xmat2):
          return False
      if wp.static(opt_broadphase_filter & BroadphaseFilter.OBB):
        if not _obb_filter(center1, center2, size1, size2, margin1, margin2, xpos1, xpos2, xmat1, xmat2):
          return False

    return True

  return func


@wp.func
def _add_geom_pair(
  # Model:
  geom_type: wp.array(dtype=int),
  nxn_pairid: wp.array(dtype=wp.vec2i),
  # Data in:
  naconmax_in: int,
  # In:
  geom1: int,
  geom2: int,
  worldid: int,
  nxnid: int,
  # Data out:
  ncollision_out: wp.array(dtype=int),
  # Out:
  collision_pair_out: wp.array(dtype=wp.vec2i),
  collision_pairid_out: wp.array(dtype=wp.vec2i),
  collision_worldid_out: wp.array(dtype=int),
):
  pairid = wp.atomic_add(ncollision_out, 0, 1)

  if pairid >= naconmax_in:
    return

  type1 = geom_type[geom1]
  type2 = geom_type[geom2]

  if type1 > type2:
    pair = wp.vec2i(geom2, geom1)
  else:
    pair = wp.vec2i(geom1, geom2)

  collision_pair_out[pairid] = pair
  collision_pairid_out[pairid] = nxn_pairid[nxnid]
  collision_worldid_out[pairid] = worldid


@wp.func
def _binary_search(values: wp.array(dtype=Any), value: Any, lower: int, upper: int) -> int:
  while lower < upper:
    mid = (lower + upper) >> 1
    if values[mid] > value:
      upper = mid
    else:
      lower = mid + 1

  return upper


def _sap_project(opt_broadphase: int):
  @wp.kernel(module="unique", enable_backward=False)
  def sap_project(
    # Model:
    ngeom: int,
    geom_rbound: wp.array2d(dtype=float),
    geom_margin: wp.array2d(dtype=float),
    # Data in:
    geom_xpos_in: wp.array2d(dtype=wp.vec3),
    nworld_in: int,
    # In:
    direction_in: wp.vec3,
    # Out:
    projection_lower_out: wp.array2d(dtype=float),
    projection_upper_out: wp.array2d(dtype=float),
    sort_index_out: wp.array2d(dtype=int),
    segmented_index_out: wp.array(dtype=int),
  ):
    worldid, geomid = wp.tid()

    xpos = geom_xpos_in[worldid, geomid]
    rbound = geom_rbound[worldid % geom_rbound.shape[0], geomid]

    if rbound == 0.0:
      # geom is a plane
      rbound = MJ_MAXVAL

    radius = rbound + geom_margin[worldid % geom_margin.shape[0], geomid]
    center = wp.dot(direction_in, xpos)

    sort_index_out[worldid, geomid] = geomid
    if not wp.isnan(center):
      projection_lower_out[worldid, geomid] = center - radius
      projection_upper_out[worldid, geomid] = center + radius
    else:
      projection_lower_out[worldid, geomid] = MJ_MAXVAL
      projection_upper_out[worldid, geomid] = MJ_MAXVAL

    if wp.static(opt_broadphase == BroadphaseType.SAP_SEGMENTED):
      if geomid == 0:
        segmented_index_out[worldid] = worldid * ngeom
        if worldid == nworld_in - 1:
          segmented_index_out[nworld_in] = nworld_in * ngeom

  return sap_project


@wp.kernel
def _sap_range(
  # Model:
  ngeom: int,
  # In:
  projection_lower_in: wp.array2d(dtype=float),
  projection_upper_in: wp.array2d(dtype=float),
  sort_index_in: wp.array2d(dtype=int),
  # Out:
  range_out: wp.array2d(dtype=int),
):
  worldid, geomid = wp.tid()

  # current bounding geom
  idx = sort_index_in[worldid, geomid]

  upper = projection_upper_in[worldid, idx]

  limit = _binary_search(projection_lower_in[worldid], upper, geomid + 1, ngeom)
  limit = wp.min(ngeom - 1, limit)

  # range of geoms for the sweep and prune process
  range_out[worldid, geomid] = limit - geomid


@cache_kernel
def _sap_broadphase(opt_broadphase_filter: int, ngeom_aabb: int, ngeom_rbound: int, ngeom_margin: int):
  @wp.kernel(module="unique", enable_backward=False)
  def kernel(
    # Model:
    ngeom: int,
    geom_type: wp.array(dtype=int),
    geom_aabb: wp.array3d(dtype=wp.vec3),
    geom_rbound: wp.array2d(dtype=float),
    geom_margin: wp.array2d(dtype=float),
    nxn_pairid: wp.array(dtype=wp.vec2i),
    # Data in:
    geom_xpos_in: wp.array2d(dtype=wp.vec3),
    geom_xmat_in: wp.array2d(dtype=wp.mat33),
    nworld_in: int,
    naconmax_in: int,
    # In:
    sort_index_in: wp.array2d(dtype=int),
    cumulative_sum_in: wp.array(dtype=int),
    nsweep_in: int,
    # Data out:
    ncollision_out: wp.array(dtype=int),
    # Out:
    collision_pair_out: wp.array(dtype=wp.vec2i),
    collision_pairid_out: wp.array(dtype=wp.vec2i),
    collision_worldid_out: wp.array(dtype=int),
  ):
    worldgeomid = wp.tid()

    nworldgeom = nworld_in * ngeom
    nworkpackages = cumulative_sum_in[nworldgeom - 1]

    while worldgeomid < nworkpackages:
      # binary search to find current and next geom pair indices
      i = _binary_search(cumulative_sum_in, worldgeomid, 0, nworldgeom)
      j = i + worldgeomid + 1

      if i > 0:
        j -= cumulative_sum_in[i - 1]

      worldid = i // ngeom
      i = i % ngeom
      j = j % ngeom

      # get geom indices and swap if necessary
      geom1 = sort_index_in[worldid, i]
      geom2 = sort_index_in[worldid, j]

      # find linear index of (geom1, geom2) in upper triangular nxn_pairid
      if geom2 < geom1:
        idx = upper_tri_index(ngeom, geom2, geom1)
      else:
        idx = upper_tri_index(ngeom, geom1, geom2)

      worldgeomid += nsweep_in
      pairid = nxn_pairid[idx]
      if pairid[0] < -1 and pairid[1] < 0:
        continue

      if (
        wp.static(_broadphase_filter(opt_broadphase_filter, ngeom_aabb, ngeom_rbound, ngeom_margin))(
          geom_aabb, geom_rbound, geom_margin, geom_xpos_in, geom_xmat_in, geom1, geom2, worldid
        )
        or pairid[1] >= 0
      ):
        _add_geom_pair(
          geom_type,
          nxn_pairid,
          naconmax_in,
          geom1,
          geom2,
          worldid,
          idx,
          ncollision_out,
          collision_pair_out,
          collision_pairid_out,
          collision_worldid_out,
        )

  return kernel


def _segmented_sort(tile_size: int):
  @wp.kernel(module="unique")
  def segmented_sort(
    # In:
    projection_lower_in: wp.array2d(dtype=float),
    sort_index_in: wp.array2d(dtype=int),
    # Out:
    projection_lower_out: wp.array2d(dtype=float),
    sort_index_out: wp.array2d(dtype=int),
  ):
    worldid = wp.tid()

    # Load input into shared memory
    keys = wp.tile_load(projection_lower_in[worldid], shape=tile_size, storage="shared")
    values = wp.tile_load(sort_index_in[worldid], shape=tile_size, storage="shared")

    # Perform in-place sorting
    wp.tile_sort(keys, values)

    # Store sorted shared memory into output arrays
    wp.tile_store(projection_lower_out[worldid], keys)
    wp.tile_store(sort_index_out[worldid], values)

  return segmented_sort


@event_scope
def sap_broadphase(m: Model, d: Data, ctx: CollisionContext):
  """Runs broadphase collision detection using a sweep-and-prune (SAP) algorithm.

  This method is more efficient than the N-squared approach for large numbers of
  objects. It works by projecting the bounding spheres of all geoms onto a
  single axis and sorting them. It then sweeps along the axis, only checking
  for overlaps between geoms whose projections are close to each other.

  For each potentially colliding pair identified by the sweep, a more precise
  bounding sphere check is performed. If this check passes, the pair is added
  to the collision arrays in `d` for the narrowphase stage.

  Two sorting strategies are supported, controlled by `m.opt.broadphase`

  - `SAP_TILE`: Uses a tile-based sort.
  - `SAP_SEGMENTED`: Uses a segmented sort.
  """
  nworldgeom = d.nworld * m.ngeom

  # TODO(team): direction

  # random fixed direction
  direction = wp.vec3(0.5935, 0.7790, 0.1235)
  direction = wp.normalize(direction)

  projection_lower = wp.empty((d.nworld, m.ngeom, 2), dtype=float)
  projection_upper = wp.empty((d.nworld, m.ngeom), dtype=float)
  sort_index = wp.empty((d.nworld, m.ngeom, 2), dtype=int)
  range_ = wp.empty((d.nworld, m.ngeom), dtype=int)
  cumulative_sum = wp.empty((d.nworld, m.ngeom), dtype=int)
  segmented_index = wp.empty(d.nworld + 1 if m.opt.broadphase == BroadphaseType.SAP_SEGMENTED else 0, dtype=int)

  wp.launch(
    kernel=_sap_project(m.opt.broadphase),
    dim=(d.nworld, m.ngeom),
    inputs=[m.ngeom, m.geom_rbound, m.geom_margin, d.geom_xpos, d.nworld, direction],
    outputs=[
      projection_lower.reshape((-1, m.ngeom)),
      projection_upper,
      sort_index.reshape((-1, m.ngeom)),
      segmented_index,
    ],
  )

  if m.opt.broadphase == BroadphaseType.SAP_TILE:
    wp.launch_tiled(
      kernel=_segmented_sort(m.ngeom),
      dim=d.nworld,
      inputs=[projection_lower.reshape((-1, m.ngeom)), sort_index.reshape((-1, m.ngeom))],
      outputs=[projection_lower.reshape((-1, m.ngeom)), sort_index.reshape((-1, m.ngeom))],
      block_dim=m.block_dim.segmented_sort,
    )
  else:
    wp.utils.segmented_sort_pairs(
      projection_lower.reshape((-1, m.ngeom)), sort_index.reshape((-1, m.ngeom)), nworldgeom, segmented_index
    )

  wp.launch(
    kernel=_sap_range,
    dim=(d.nworld, m.ngeom),
    inputs=[m.ngeom, projection_lower.reshape((-1, m.ngeom)), projection_upper, sort_index.reshape((-1, m.ngeom))],
    outputs=[range_],
  )

  # scan is used for load balancing among the threads
  wp.utils.array_scan(range_.reshape(-1), cumulative_sum.reshape(-1), True)

  # estimate number of overlap checks
  # assumes each geom has 5 other geoms (batched over all worlds)
  nsweep = 5 * nworldgeom
  wp.launch(
    kernel=_sap_broadphase(m.opt.broadphase_filter, m.geom_aabb.shape[0], m.geom_rbound.shape[0], m.geom_margin.shape[0]),
    dim=nsweep,
    inputs=[
      m.ngeom,
      m.geom_type,
      m.geom_aabb,
      m.geom_rbound,
      m.geom_margin,
      m.nxn_pairid,
      d.geom_xpos,
      d.geom_xmat,
      d.nworld,
      d.naconmax,
      sort_index.reshape((-1, m.ngeom)),
      cumulative_sum.reshape(-1),
      nsweep,
    ],
    outputs=[d.ncollision, ctx.collision_pair, ctx.collision_pairid, ctx.collision_worldid],
  )


@cache_kernel
def _nxn_broadphase(opt_broadphase_filter: int, ngeom_aabb: int, ngeom_rbound: int, ngeom_margin: int):
  @wp.kernel(module="unique", enable_backward=False)
  def kernel(
    # Model:
    geom_type: wp.array(dtype=int),
    geom_aabb: wp.array3d(dtype=wp.vec3),
    geom_rbound: wp.array2d(dtype=float),
    geom_margin: wp.array2d(dtype=float),
    nxn_geom_pair: wp.array(dtype=wp.vec2i),
    nxn_pairid: wp.array(dtype=wp.vec2i),
    # Data in:
    geom_xpos_in: wp.array2d(dtype=wp.vec3),
    geom_xmat_in: wp.array2d(dtype=wp.mat33),
    naconmax_in: int,
    # Data out:
    ncollision_out: wp.array(dtype=int),
    # Out:
    collision_pair_out: wp.array(dtype=wp.vec2i),
    collision_pairid_out: wp.array(dtype=wp.vec2i),
    collision_worldid_out: wp.array(dtype=int),
  ):
    worldid, elementid = wp.tid()

    geom = nxn_geom_pair[elementid]
    geom1 = geom[0]
    geom2 = geom[1]

    if (
      wp.static(_broadphase_filter(opt_broadphase_filter, ngeom_aabb, ngeom_rbound, ngeom_margin))(
        geom_aabb, geom_rbound, geom_margin, geom_xpos_in, geom_xmat_in, geom1, geom2, worldid
      )
      or nxn_pairid[elementid][1] >= 0
    ):
      _add_geom_pair(
        geom_type,
        nxn_pairid,
        naconmax_in,
        geom1,
        geom2,
        worldid,
        elementid,
        ncollision_out,
        collision_pair_out,
        collision_pairid_out,
        collision_worldid_out,
      )

  return kernel


@event_scope
def nxn_broadphase(m: Model, d: Data, ctx: CollisionContext):
  """Runs broadphase collision detection using a brute-force N-squared approach.

  This function iterates through a pre-filtered list of all possible geometry pairs and
  performs a quick bounding sphere check to identify potential collisions.

  For each pair that passes the sphere check, it populates the collision arrays in `d`
  (`d.collision_pair`, `d.collision_pairid`, etc.), which are then consumed by the
  narrowphase.

  The initial list of pairs is filtered at model creation time to exclude pairs based on
  `contype`/`conaffinity`, parent-child relationships, and explicit `<exclude>` tags.
  """
  wp.launch(
    _nxn_broadphase(m.opt.broadphase_filter, m.geom_aabb.shape[0], m.geom_rbound.shape[0], m.geom_margin.shape[0]),
    dim=(d.nworld, m.nxn_geom_pair_filtered.shape[0]),
    inputs=[
      m.geom_type,
      m.geom_aabb,
      m.geom_rbound,
      m.geom_margin,
      m.nxn_geom_pair_filtered,
      m.nxn_pairid_filtered,
      d.geom_xpos,
      d.geom_xmat,
      d.naconmax,
    ],
    outputs=[
      d.ncollision,
      ctx.collision_pair,
      ctx.collision_pairid,
      ctx.collision_worldid,
    ],
  )


def _narrowphase(m: Model, d: Data, ctx: CollisionContext):
  collision_table = MJ_COLLISION_TABLE
  if m.opt.disableflags & DisableBit.NATIVECCD:
    collision_table[(GeomType.BOX, GeomType.BOX)] = CollisionType.PRIMITIVE

  convex_pairs = [key for key, value in collision_table.items() if value == CollisionType.CONVEX]
  primitive_pairs = [key for key, value in collision_table.items() if value == CollisionType.PRIMITIVE]

  # TODO(team): we should reject far-away contacts in the narrowphase instead of constraint
  #             partitioning because we can move some pressure of the atomics
  convex_narrowphase(m, d, ctx, convex_pairs)
  primitive_narrowphase(m, d, ctx, primitive_pairs)

  if m.has_sdf_geom:
    sdf_narrowphase(m, d, ctx)


@event_scope
def collision(m: Model, d: Data):
  """Runs the full collision detection pipeline.

  This function orchestrates the broadphase and narrowphase collision detection stages. It
  first identifies potential collision pairs using a broadphase algorithm (either N-squared
  or Sweep-and-Prune, based on `m.opt.broadphase`). Then, for each potential pair, it
  performs narrowphase collision detection to compute detailed contact information like
  distance, position, and frame.

  The results are used to populate the `d.contact` array, and the total number of contacts
  is stored in `d.nacon`.  If `d.nacon` is larger than `d.naconmax` then an overflow has
  occurred and the remaining contacts will be skipped.  If this happens, raise the `nconmax`
  parameter in `io.make_data` or `io.put_data`.

  This function will do nothing except zero out arrays if collision detection is disabled
  via `m.opt.disableflags` or if `d.nacon` is 0.
  """
  if d.naconmax == 0 or m.opt.disableflags & (DisableBit.CONSTRAINT | DisableBit.CONTACT):
    d.nacon.zero_()
    return

  ctx = create_collision_context(d.naconmax)

  # zero counters
  wp.launch(_zero_nacon_ncollision, dim=1, outputs=[d.nacon, d.ncollision])

  if m.opt.broadphase == BroadphaseType.NXN:
    nxn_broadphase(m, d, ctx)
  else:
    sap_broadphase(m, d, ctx)

  _narrowphase(m, d, ctx)



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import warp as wp
from absl.testing import absltest
from absl.testing import parameterized

from mujoco_warp import Data
from mujoco_warp import GeomType
from mujoco_warp import Model
from mujoco_warp import test_data
from mujoco_warp._src.collision_gjk import ccd
from mujoco_warp._src.collision_gjk import multicontact
from mujoco_warp._src.collision_gjk import support
from mujoco_warp._src.collision_primitive import Geom
from mujoco_warp._src.types import MJ_MAX_EPAFACES
from mujoco_warp._src.types import MJ_MAX_EPAHORIZON
from mujoco_warp._src.types import mat63


def _geom_dist(
  m: Model,
  d: Data,
  gid1: int,
  gid2: int,
  multiccd=False,
  margin=0.0,
  pos1: wp.vec3 | None = None,
  pos2: wp.vec3 | None = None,
  mat1: wp.mat33 | None = None,
  mat2: wp.mat33 | None = None,
):
  # we run multiccd on static scenes so these need to be initialized
  nmaxpolygon = 10 if multiccd else 0
  nmaxmeshdeg = 10 if multiccd else 0
  epa_vert = wp.empty(10 + 2 * m.opt.ccd_iterations, dtype=wp.vec3)
  epa_vert_index = wp.empty(10 + 2 * m.opt.ccd_iterations, dtype=int)
  epa_face = wp.empty(6 + MJ_MAX_EPAFACES * m.opt.ccd_iterations, dtype=int)
  epa_pr = wp.empty(6 + MJ_MAX_EPAFACES * m.opt.ccd_iterations, dtype=wp.vec3)
  epa_norm2 = wp.empty(6 + MJ_MAX_EPAFACES * m.opt.ccd_iterations, dtype=float)
  epa_horizon = wp.empty(MJ_MAX_EPAHORIZON, dtype=int)
  multiccd_polygon = wp.empty(2 * nmaxpolygon, dtype=wp.vec3)
  multiccd_clipped = wp.empty(2 * nmaxpolygon, dtype=wp.vec3)
  multiccd_pnormal = wp.empty(nmaxpolygon, dtype=wp.vec3)
  multiccd_pdist = wp.empty(nmaxpolygon, dtype=float)
  multiccd_idx1 = wp.empty(nmaxmeshdeg, dtype=int)
  multiccd_idx2 = wp.empty(nmaxmeshdeg, dtype=int)
  multiccd_n1 = wp.empty(nmaxmeshdeg, dtype=wp.vec3)
  multiccd_n2 = wp.empty(nmaxmeshdeg, dtype=wp.vec3)
  multiccd_endvert = wp.empty(nmaxmeshdeg, dtype=wp.vec3)
  multiccd_face1 = wp.empty(nmaxpolygon, dtype=wp.vec3)
  multiccd_face2 = wp.empty(nmaxpolygon, dtype=wp.vec3)

  @wp.kernel(module="unique", enable_backward=False)
  def _ccd_kernel(
    # Model:
    geom_type: wp.array(dtype=int),
    geom_dataid: wp.array(dtype=int),
    geom_size: wp.array2d(dtype=wp.vec3),
    mesh_vertadr: wp.array(dtype=int),
    mesh_vertnum: wp.array(dtype=int),
    mesh_vert: wp.array(dtype=wp.vec3),
    mesh_polynum: wp.array(dtype=int),
    mesh_polyadr: wp.array(dtype=int),
    mesh_polynormal: wp.array(dtype=wp.vec3),
    mesh_polyvertadr: wp.array(dtype=int),
    mesh_polyvertnum: wp.array(dtype=int),
    mesh_polyvert: wp.array(dtype=int),
    mesh_polymapadr: wp.array(dtype=int),
    mesh_polymapnum: wp.array(dtype=int),
    mesh_polymap: wp.array(dtype=int),
    # Data in:
    geom_xpos_in: wp.array2d(dtype=wp.vec3),
    geom_xmat_in: wp.array2d(dtype=wp.mat33),
    # In:
    gid1: int,
    gid2: int,
    iterations: int,
    tolerance: wp.array(dtype=float),
    vert: wp.array(dtype=wp.vec3),
    vert_index: wp.array(dtype=int),
    face: wp.array(dtype=int),
    face_pr: wp.array(dtype=wp.vec3),
    face_norm2: wp.array(dtype=float),
    horizon: wp.array(dtype=int),
    polygon: wp.array(dtype=wp.vec3),
    clipped: wp.array(dtype=wp.vec3),
    pnormal: wp.array(dtype=wp.vec3),
    pdist: wp.array(dtype=float),
    idx1: wp.array(dtype=int),
    idx2: wp.array(dtype=int),
    n1: wp.array(dtype=wp.vec3),
    n2: wp.array(dtype=wp.vec3),
    endvert: wp.array(dtype=wp.vec3),
    face1: wp.array(dtype=wp.vec3),
    face2: wp.array(dtype=wp.vec3),
    # Out:
    dist_out: wp.array(dtype=float),
    ncon_out: wp.array(dtype=int),
    pos_out: wp.array(dtype=wp.vec3),
  ):
    geom1 = Geom()
    geom1.index = -1
    geomtype1 = geom_type[gid1]
    if wp.static(pos1 == None):
      geom1.pos = geom_xpos_in[0, gid1]
    else:
      geom1.pos = pos1
    if wp.static(mat1 == None):
      geom1.rot = geom_xmat_in[0, gid1]
    else:
      geom1.rot = mat1
    geom1.size = geom_size[0, gid1]
    geom1.margin = margin
    geom1.graphadr = -1
    geom1.mesh_polyadr = -1

    if geom_dataid[gid1] >= 0 and geom_type[gid1] == GeomType.MESH:
      dataid = geom_dataid[gid1]
      geom1.vertadr = mesh_vertadr[dataid]
      geom1.vertnum = mesh_vertnum[dataid]
      geom1.mesh_polynum = mesh_polynum[dataid]
      geom1.mesh_polyadr = mesh_polyadr[dataid]
      geom1.vert = mesh_vert
      geom1.mesh_polynormal = mesh_polynormal
      geom1.mesh_polyvertadr = mesh_polyvertadr
      geom1.mesh_polyvertnum = mesh_polyvertnum
      geom1.mesh_polyvert = mesh_polyvert
      geom1.mesh_polymapadr = mesh_polymapadr
      geom1.mesh_polymapnum = mesh_polymapnum
      geom1.mesh_polymap = mesh_polymap

    geom2 = Geom()
    geom2.index = -1
    geomtype2 = geom_type[gid2]
    if wp.static(pos2 == None):
      geom2.pos = geom_xpos_in[0, gid2]
    else:
      geom2.pos = pos2
    if wp.static(mat2 == None):
      geom2.rot = geom_xmat_in[0, gid2]
    else:
      geom2.rot = mat2
    geom2.size = geom_size[0, gid2]
    geom2.margin = margin
    geom2.graphadr = -1
    geom2.mesh_polyadr = -1

    if geom_dataid[gid2] >= 0 and geom_type[gid2] == GeomType.MESH:
      dataid = geom_dataid[gid2]
      geom2.vertadr = mesh_vertadr[dataid]
      geom2.vertnum = mesh_vertnum[dataid]
      geom2.mesh_polynum = mesh_polynum[dataid]
      geom2.mesh_polyadr = mesh_polyadr[dataid]
      geom2.vert = mesh_vert
      geom2.mesh_polynormal = mesh_polynormal
      geom2.mesh_polyvertadr = mesh_polyvertadr
      geom2.mesh_polyvertnum = mesh_polyvertnum
      geom2.mesh_polyvert = mesh_polyvert
      geom2.mesh_polymapadr = mesh_polymapadr
      geom2.mesh_polymapnum = mesh_polymapnum
      geom2.mesh_polymap = mesh_polymap

    (
      dist,
      ncon,
      x1,
      x2,
      idx,
    ) = ccd(
      tolerance[0],
      1.0e30,
      iterations,
      iterations,
      geom1,
      geom2,
      geomtype1,
      geomtype2,
      geom1.pos,
      geom2.pos,
      vert,
      vert_index,
      face,
      face_pr,
      face_norm2,
      horizon,
    )

    if wp.static(multiccd):
      ncon, _, _ = multicontact(
        polygon,
        clipped,
        pnormal,
        pdist,
        idx1,
        idx2,
        n1,
        n2,
        endvert,
        face1,
        face2,
        vert,
        vert_index,
        face[idx],
        x1,
        x2,
        geom1,
        geom2,
        geomtype1,
        geomtype2,
      )

    dist_out[0] = dist
    ncon_out[0] = ncon
    pos_out[0] = x1
    pos_out[1] = x2

  dist_out = wp.array(shape=(1,), dtype=float)
  ncon_out = wp.array(shape=(1,), dtype=int)
  pos_out = wp.array(shape=(2,), dtype=wp.vec3)
  wp.launch(
    _ccd_kernel,
    dim=1,
    inputs=[
      m.geom_type,
      m.geom_dataid,
      m.geom_size,
      m.mesh_vertadr,
      m.mesh_vertnum,
      m.mesh_vert,
      m.mesh_polynum,
      m.mesh_polyadr,
      m.mesh_polynormal,
      m.mesh_polyvertadr,
      m.mesh_polyvertnum,
      m.mesh_polyvert,
      m.mesh_polymapadr,
      m.mesh_polymapnum,
      m.mesh_polymap,
      d.geom_xpos,
      d.geom_xmat,
      gid1,
      gid2,
      m.opt.ccd_iterations,
      m.opt.ccd_tolerance,
      epa_vert,
      epa_vert_index,
      epa_face,
      epa_pr,
      epa_norm2,
      epa_horizon,
      multiccd_polygon,
      multiccd_clipped,
      multiccd_pnormal,
      multiccd_pdist,
      multiccd_idx1,
      multiccd_idx2,
      multiccd_n1,
      multiccd_n2,
      multiccd_endvert,
      multiccd_face1,
      multiccd_face2,
    ],
    outputs=[
      dist_out,
      ncon_out,
      pos_out,
    ],
  )
  return dist_out.numpy()[0], ncon_out.numpy()[0], pos_out.numpy()[0], pos_out.numpy()[1]


class GJKTest(parameterized.TestCase):
  """Tests for GJK/EPA."""

  def test_spheres_distance(self):
    """Test distance between two spheres."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <worldbody>
          <geom name="geom1" type="sphere" pos="-1.5 0 0" size="1"/>
          <geom name="geom2" type="sphere" pos="1.5 0 0" size="1"/>
        </worldbody>
       </mujoco>
       """
    )

    dist, _, x1, x2 = _geom_dist(m, d, 0, 1)
    self.assertEqual(1.0, dist)
    self.assertEqual(-0.5, x1[0])
    self.assertEqual(0.5, x2[0])

  def test_spheres_touching(self):
    """Test two touching spheres have zero distance."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <worldbody>
          <geom type="sphere" pos="-1 0 0" size="1"/>
          <geom type="sphere" pos="1 0 0" size="1"/>
        </worldbody>
       </mujoco>
       """
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1)
    self.assertEqual(0.0, dist)

  def test_box_mesh_distance(self):
    """Test distance between a mesh and box."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco model="MuJoCo Model">
        <asset>
          <mesh name="smallbox" scale="0.1 0.1 0.1"
                vertex="-1 -1 -1
                         1 -1 -1
                         1  1 -1
                         1  1  1
                         1 -1  1
                        -1  1 -1
                        -1  1  1
                        -1 -1  1"/>
         </asset>
         <worldbody>
           <geom pos="0 0 .90" type="box" size="0.5 0.5 0.1"/>
           <geom pos="0 0 1.2" type="mesh" mesh="smallbox"/>
          </worldbody>
       </mujoco>
       """
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1)
    self.assertAlmostEqual(0.1, dist)

  def test_sphere_sphere_contact(self):
    """Test penetration depth between two spheres."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <worldbody>
          <geom type="sphere" pos="-1 0 0" size="3"/>
          <geom type="sphere" pos=" 3 0 0" size="3"/>
        </worldbody>
      </mujoco>
      """
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1, 0)
    self.assertAlmostEqual(-2, dist)

  def test_box_box_contact(self):
    """Test penetration between two boxes."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <worldbody>
          <geom type="box" pos="-1 0 0" size="2.5 2.5 2.5"/>
          <geom type="box" pos="1.5 0 0" size="1 1 1"/>
        </worldbody>
      </mujoco>
      """
    )
    dist, _, x1, x2 = _geom_dist(m, d, 0, 1)
    diff = x1 - x2
    normal = diff / np.linalg.norm(diff)

    self.assertAlmostEqual(-1, dist)
    self.assertAlmostEqual(normal[0], 1)
    self.assertAlmostEqual(normal[1], 0)
    self.assertAlmostEqual(normal[2], 0)

  def test_mesh_mesh_contact(self):
    """Test penetration between two meshes."""
    _, _, m, d = test_data.fixture(
      xml="""
    <mujoco>
      <asset>
        <mesh name="box" scale=".5 .5 .1"
              vertex="-1 -1 -1
                       1 -1 -1
                       1  1 -1
                       1  1  1
                       1 -1  1
                      -1  1 -1
                      -1  1  1
                      -1 -1  1"/>
        <mesh name="smallbox" scale=".1 .1 .1"
              vertex="-1 -1 -1
                       1 -1 -1
                       1  1 -1
                       1  1  1
                       1 -1  1
                      -1  1 -1
                      -1  1  1
                      -1 -1  1"/>
      </asset>

      <worldbody>
        <geom pos="0 0 .09" type="mesh" mesh="smallbox"/>
        <geom pos="0 0 -.1" type="mesh" mesh="box"/>
      </worldbody>
    </mujoco>
    """
    )
    dist, _, _, _ = _geom_dist(m, d, 0, 1)
    self.assertAlmostEqual(-0.01, dist)

  def test_cylinder_cylinder_contact(self):
    """Test penetration between two cylinder."""
    _, _, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <worldbody>
          <geom pos="0 0 0" type="cylinder" size="1 .5"/>
          <geom pos="1.999 0 0" type="cylinder" size="1 .5"/>
        </worldbody>
      </mujoco>
    """
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1)
    self.assertAlmostEqual(-0.001, dist)

  def test_box_edge(self):
    """Test box edge."""
    _, _, m, d = test_data.fixture(
      xml="""
    <mujoco>
      <worldbody>
        <geom pos="0 0 2" type="box" name="box2" size="1 1 1"/>
        <geom pos="0 0 4.4" euler="0 90 40" type="box" name="box3" size="1 1 1"/>
      </worldbody>
    </mujoco>"""
    )
    _, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True)
    self.assertEqual(ncon, 2)

  def test_box_box_ccd(self):
    """Test box box."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
           <geom name="geom1" type="box" pos="0 0 1.9" size="1 1 1"/>
           <geom name="geom2" type="box" pos="0 0 0" size="10 10 1"/>
         </worldbody>
       </mujoco>
       """
    )
    _, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True)
    self.assertEqual(ncon, 4)

  def test_mesh_mesh_ccd(self):
    """Test mesh-mesh multiccd."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <asset>
           <mesh name="smallbox"
                 vertex="-1 -1 -1 1 -1 -1 1 1 -1 1 1 1 1 -1 1 -1 1 -1 -1 1 1 -1 -1 1"/>
         </asset>
         <worldbody>
           <geom pos="0 0 2" type="mesh" name="box1" mesh="smallbox"/>
          <geom pos="0 1 3.99" euler="0 0 40" type="mesh" name="box2" mesh="smallbox"/>
         </worldbody>
       </mujoco>
       """
    )

    _, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True)
    self.assertEqual(ncon, 4)

  def test_box_box_ccd2(self):
    """Test box-box multiccd 2."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
           <geom size="1 1 1" pos="0 0 2" type="box"/>
          <geom size="1 1 1" pos="0 1 3.99" euler="0 0 40" type="box"/>
         </worldbody>
       </mujoco>
       """
    )

    _, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True)
    self.assertEqual(ncon, 4)

  def test_sphere_mesh_margin(self):
    """Test sphere-mesh margin."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <asset>
           <mesh name="box" scale=".2 .2 .2"
                 vertex="-1 -1 -1 1 -1 -1 1 1 -1 1 1 1 1 -1 1 -1 1 -1 -1 1 1 -1 -1 1"/>
         </asset>
         <worldbody>
           <geom type="sphere" pos="0 0 .349" size=".1"/>
           <geom type="mesh" mesh="box"/>
         </worldbody>
       </mujoco>
       """
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1, multiccd=False, margin=0.05)
    self.assertAlmostEqual(dist, -0.001)

  def test_cylinder_box(self):
    """Test cylinder box collision."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
           <geom type="box" size="1 1 0.1"/>
           <geom type="cylinder" size=".1 .2 .3"/>
         </worldbody>
       </mujoco>
       """,
      overrides=["opt.ccd_iterations=50"],
    )

    pos = wp.vec3(0.00015228791744448245, -0.00074981129728257656, 0.29839199781417846680)
    rot = wp.mat33(
      0.99996972084045410156,
      0.00776371126994490623,
      -0.00043433305108919740,
      -0.00776385562494397163,
      0.99996984004974365234,
      -0.00033095158869400620,
      0.00043175052269361913,
      0.00033431366318836808,
      0.99999988079071044922,
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1, pos2=pos, mat2=rot)
    self.assertAlmostEqual(dist, -0.0016624178339902445)

  def test_box_box_float(self):
    """Test box-box under float32."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
          <geom size=".025 .025 .025" type="box"/>
          <geom size=".025 .025 .025" type="box"/>
         </worldbody>
       </mujoco>
       """
    )

    pos1 = wp.vec3(-0.17624500393867492676, -0.12375499308109283447, 0.12499777972698211670)
    rot1 = wp.mat33(
      1.00000000000000000000,
      -0.00000000184385418045,
      -0.00000025833372774287,
      0.00000000184391857339,
      1.00000000000000000000,
      0.00000024928382913458,
      0.00000025833372774287,
      -0.00000024928382913458,
      1.0000000000000000000,
    )

    pos2 = wp.vec3(-0.17624500393867492676, -0.12375499308109283447, 0.17499557137489318848)
    rot2 = wp.mat33(
      1.00000000000000000000,
      -0.00000000184292525685,
      0.00000012980596864054,
      0.00000000184294413064,
      1.00000000000000000000,
      -0.00000014602545661546,
      -0.00000012980596864054,
      0.00000014602545661546,
      1.00000000000000000000,
    )

    dist, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=False, pos1=pos1, mat1=rot1, pos2=pos2, mat2=rot2)
    self.assertEqual(ncon, 1)
    self.assertLess(dist, 0.0001)  # real depth is ~ 2E-6

  def test_box_box_horizon(self):
    """Test box-box with EPA horizon with 13 edges."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
          <geom size=".025 .025 .025" type="box"/>
          <geom size=".025 .025 .025" type="box"/>
         </worldbody>
       </mujoco>
       """
    )

    pos1 = wp.vec3(0.065118454396725, -0.125125020742416, 0.124963559210300)
    rot1 = wp.mat33(
      0.996357858181000,
      0.085266821086407,
      -0.000942531623878,
      -0.085266284644604,
      0.996358215808868,
      0.000591202871874,
      0.000989508931525,
      -0.000508683384396,
      0.999999582767487,
    )

    pos2 = wp.vec3(0.065104484558105, -0.124979749321938, 0.174992129206657)
    rot2 = wp.mat33(
      0.996556758880615,
      -0.082913912832737,
      -0.000453041866422,
      0.082915119826794,
      0.996536433696747,
      0.006357696373016,
      -0.000075668765930,
      -0.006373368669301,
      0.999979794025421,
    )

    dist, _, _, _ = _geom_dist(m, d, 0, 1, multiccd=False, pos1=pos1, mat1=rot1, pos2=pos2, mat2=rot2)
    self.assertAlmostEqual(dist, -0.00011578822, 6)  # dist = -0.00011579410621457821 - MJC 64 bit precision

  def test_box_box_rotation(self):
    """Test box-box with slight rotation which should give 4 contacts."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
          <geom size=".025 .025 .025" type="box"/>
          <geom size=".025 .025 .025" type="box"/>
         </worldbody>
       </mujoco>
       """
    )

    pos1 = wp.vec3(
      0.015344001352787,
      -0.195344015955925,
      0.174637570977211,
    )
    rot1 = wp.mat33(
      1.000000000000000,
      0.000000000029901,
      0.000004057303613,
      -0.000000000062404,
      1.000000000000000,
      0.000008010840247,
      -0.000004057303613,
      -0.000008010840247,
      1.000000000000000,
    )

    pos2 = wp.vec3(
      0.015344001352787,
      -0.195344015955925,
      0.224056228995323,
    )
    rot2 = wp.mat33(
      1.000000000000000,
      0.000000000029692,
      -0.000003355821491,
      -0.000000000057016,
      1.000000000000000,
      -0.000008142159459,
      0.000003355821491,
      0.000008142159459,
      1.000000000000000,
    )

    _, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True, pos1=pos1, mat1=rot1, pos2=pos2, mat2=rot2)
    self.assertEqual(ncon, 4)

  def test_box_box_diagonal(self):
    """Test box-box where multiccd has a diagonal edge as a face."""
    _, _, m, d = test_data.fixture(
      xml="""
       <mujoco>
         <worldbody>
          <geom size="0.50 0.50 0.10" type="box"/>
          <geom size=".025 .025 .025" type="box"/>
         </worldbody>
       </mujoco>
       """
    )

    pos2 = wp.vec3(
      0.135535001754761,
      -0.195535004138947,
      0.124984227120876,
    )
    rot2 = wp.mat33(
      1.000000000000000,
      0.000000000048563,
      -0.000000135524601,
      -0.000000000048577,
      1.000000000000000,
      -0.000000103374248,
      0.000000135524601,
      0.000000103374248,
      1.000000000000000,
    )

    dist, ncon, _, _ = _geom_dist(m, d, 0, 1, multiccd=True, pos2=pos2, mat2=rot2)
    self.assertAlmostEqual(dist, -1.5778851595232846e-05)
    self.assertEqual(ncon, 4)

  @parameterized.parameters(0.0, 0.1)
  def test_hfield_support(self, margin: float):
    """Test support function for height field geoms."""
    eps = 1e-3

    # Bottom triangle (z = 0)
    # Top triangle (z = 1 + margin, following collision_convex.py pattern)
    # fmt: off
    prism = mat63(
      0.0, 0.0, 0.0,           # bottom vertex 0
      1.0, 0.0, 0.0,           # bottom vertex 1
      0.5, 1.0, 0.0,           # bottom vertex 2
      0.0, 0.0, 1.0 + margin,  # top vertex 3
      1.0, 0.0, 1.0 + margin,  # top vertex 4
      0.5, 1.0, 1.0 + margin,  # top vertex 5
    )
    # fmt: on

    @wp.kernel(module="unique", enable_backward=False)
    def _support_kernel(
      hfprism_in: mat63,
      eps_in: float,
      support_point: wp.array(dtype=wp.vec3),
    ):
      geom = Geom()
      geom.pos = wp.vec3(0.0, 0.0, 0.0)
      geom.rot = wp.identity(n=3, dtype=float)
      geom.hfprism = hfprism_in
      geom.margin = 0.0  # margin added to prism

      # Test directions with eps offsets for unique support points

      # dir = (eps, eps, 1): selects prism[5] (top, highest z, breaks tie with x,y)
      sp = support(geom, GeomType.HFIELD, wp.vec3(eps_in, eps_in, 1.0))
      support_point[0] = sp.point

      # dir = (-eps, -eps, -1): selects prism[0] (bottom, lowest z)
      sp = support(geom, GeomType.HFIELD, wp.vec3(-eps_in, -eps_in, -1.0))
      support_point[1] = sp.point

      # dir = (1, eps, eps): selects prism[4] (top, x=1, eps breaks ties)
      sp = support(geom, GeomType.HFIELD, wp.vec3(1.0, eps_in, eps_in))
      support_point[2] = sp.point

      # dir = (eps, 1, eps): selects prism[5] (top, y=1, eps breaks ties)
      sp = support(geom, GeomType.HFIELD, wp.vec3(eps_in, 1.0, eps_in))
      support_point[3] = sp.point

    support_point = wp.empty(4, dtype=wp.vec3)

    wp.launch(
      _support_kernel,
      dim=1,
      inputs=[prism, eps],
      outputs=[support_point],
    )

    result = support_point.numpy()

    # dir = (eps, eps, 1): expect prism[5] + margin offset
    np.testing.assert_allclose(result[0], prism[5], rtol=1e-5)

    # dir = (-eps, -eps, -1): expect prism[0] + margin offset
    np.testing.assert_allclose(result[1], prism[0], rtol=1e-5)

    # dir = (1, 0, eps): expect prism[4] + margin offset
    np.testing.assert_allclose(result[2], prism[4], rtol=1e-5)

    # dir = (0, 1, eps): expect prism[5] + margin offset
    np.testing.assert_allclose(result[3], prism[5], rtol=1e-5)


if __name__ == "__main__":
  wp.init()
  absltest.main()
